/*
 * synaptiks -- a touchpad control tool
 *
 *
 * Copyright (C) 2009, 2010 Sebastian Wiesner <lunaryorn@googlemail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define QT_NO_CAST_TO_ASCII 1
#include "mousedevicesmonitor.h"
#include "qudevpointer.h"
#include <KDebug>
#include <QtCore/QStringList>
#include <QtCore/QSocketNotifier>


using namespace synaptiks;


namespace synaptiks {
    class MouseDevicesMonitorPrivate {
    public:
        Q_DECLARE_PUBLIC(MouseDevicesMonitor)

        MouseDevicesMonitorPrivate(MouseDevicesMonitor *qq);

        virtual ~MouseDevicesMonitorPrivate() {};

        /**
         * Slot for monitorNotifier signals.  Reads the available device,
         * and emits a MouseDevicesMonitor signals, if the device is an
         * added or removed mouse device.  Also updates the serial number
         * and product name caches.
         */
        void _k_udevDataAvailable();

        /**
         * Find the device object for the given @p serial number.
         *
         * Device lookup is cached using serialCache.  UDev is only queried
         * on cache misses.
         *
         * @param serial the serial number
         * @return the device or 0, if there is no device for @p serial
         */
        QUDevDevicePtr findDeviceBySerial(const QByteArray &serial) const;

        /**
         * Map serial numbers to device paths to speed up device look up by
         * serial number
         */
        mutable QHash<QByteArray, QByteArray> serialCache;

        /**
         * Map serial numbers to the product name of the corresponding
         * device.
         */
        mutable QHash<QByteArray, QByteArray> productNameCache;

        /**
         * UDev library context, which is required to access the whole
         * library.
         */
        QUDevContextPtr context;

        /**
         * UDev monitor object.  Provides a file descriptor to read added or
         * removed devices from.
         */
        QUDevMonitorPtr monitor;

        /**
         * Notifier object for the monitor.  Monitors the monitor's file
         * descriptor for added or removed devices.
         */
        QSocketNotifier *monitorNotifier;

        MouseDevicesMonitor *q_ptr;
    };
}

/**
 * This code uses the serial number of input devices to clearly identify
 * mouse device.  The path is sysfs can't be used, because it depends on the
 * port, into which the device is plugged:  a single mouse device has
 * different sys paths if plugged into different ports.
 *
 * However, udev does not allow for direct property queries.  To get a
 * device object for a given serial number, all input devices are enumerated
 * and filtered by the given serial number.  Results of such lookups are
 * cached.
 */

MouseDevicesMonitorPrivate::MouseDevicesMonitorPrivate(
    MouseDevicesMonitor *qq): q_ptr(qq) {
    Q_Q(MouseDevicesMonitor);
    this->context = udev_new();
    this->monitor = udev_monitor_new_from_netlink(this->context, "udev");
    this->monitorNotifier = new QSocketNotifier(
        udev_monitor_get_fd(this->monitor), QSocketNotifier::Read, q);
    q->connect(this->monitorNotifier, SIGNAL(activated(int)),
               SLOT(_k_udevDataAvailable()));
    // we are only interested in events on input devices
    udev_monitor_filter_add_match_subsystem_devtype(
        this->monitor, "input", 0);
    udev_monitor_enable_receiving(this->monitor);
}

/**
 * @return @c true, if @p device is a mouse (not a touchpad!), @c false
 * otherwise
 */
inline bool isMouse(const QUDevDevicePtr &device) {
    // check the sys name here, because input devices may be represented by
    // more than one device object (e.g. a raw event device and a mouse
    // device).  We need to make sure, that we really got the mouse device
    // object
    QByteArray sysName = udev_device_get_sysname(device);
    QByteArray isMouse = udev_device_get_property_value(
        device, "ID_INPUT_MOUSE");
    QByteArray isTouchpad = udev_device_get_property_value(
        device, "ID_INPUT_TOUCHPAD");
    return (isMouse.toInt() && (!isTouchpad.toInt())
            && sysName.startsWith("mouse"));
}

inline QByteArray getSerial(const QUDevDevicePtr &device) {
    return udev_device_get_property_value(device, "ID_SERIAL");
}

void MouseDevicesMonitorPrivate::_k_udevDataAvailable() {
    Q_Q(MouseDevicesMonitor);
    kDebug() << "udev data available";
    QUDevDevicePtr device = udev_monitor_receive_device(this->monitor);
    if (!device)
        return;
    kDebug() << "udev device received";
    if (isMouse(device)) {
        QByteArray action = udev_device_get_action(device);
        QByteArray serial = getSerial(device);
        if (action == "add") {
            kDebug() << "mouse added:" << serial;
            QByteArray sysPath = udev_device_get_syspath(device);
            this->serialCache.insert(serial, sysPath);
            emit q->mousePlugged(QString::fromAscii(serial));
        } else if (action == "remove") {
            kDebug() << "mouse removed:" << serial;
            // remove device from all caches
            this->serialCache.remove(serial);
            this->productNameCache.remove(serial);
            emit q->mouseUnplugged(QString::fromAscii(serial));
        }
    }
}

QUDevDevicePtr MouseDevicesMonitorPrivate::findDeviceBySerial(
    const QByteArray &serial) const {
    kDebug() << "searching device for serial" << serial;
    // try to get the device from cache
    QByteArray sysPath = this->serialCache.value(serial);
    if (!sysPath.isEmpty()) {
        QUDevDevicePtr device =
            udev_device_new_from_syspath(this->context,  sysPath);
        if (device && isMouse(device)) {
            kDebug() << "found device for serial" << serial << "in cache";
            return device;
        }
    }
    // cache miss, query udev for a device with the given serial number
    QUDevEnumeratePtr enumerator = udev_enumerate_new(this->context);
    Q_ASSERT(enumerator);
    // filter for the given serial number
    udev_enumerate_add_match_subsystem(enumerator, "input");
    udev_enumerate_add_match_property(enumerator, "ID_SERIAL", serial);
    udev_enumerate_scan_devices(enumerator);
    // iterate over all matching devices
    udev_list_entry *device_entries =
        udev_enumerate_get_list_entry(enumerator);
    udev_list_entry *current_entry = 0;
    udev_list_entry_foreach(current_entry, device_entries) {
        QByteArray sysPath = udev_list_entry_get_name(current_entry);
        QUDevDevicePtr device = udev_device_new_from_syspath(
            this->context, sysPath);
        // we have a matching mouse device
        if (device && isMouse(device)) {
            kDebug() << "found device for serial:" << serial;
            // cache the result of the udev query
            this->serialCache.insert(serial, sysPath);
            return device;
        }
    }
    return 0;
}

MouseDevicesMonitor::MouseDevicesMonitor(QObject *parent):
    QObject(parent), d_ptr(new MouseDevicesMonitorPrivate(this)) {}

MouseDevicesMonitor::~MouseDevicesMonitor() {
    delete this->d_ptr;
}

QStringList MouseDevicesMonitor::pluggedMouseDevices() const {
    Q_D(const MouseDevicesMonitor);
    QStringList devices;
    kDebug() << "enumerating plugged devices";
    QUDevEnumeratePtr enumerator = udev_enumerate_new(d->context);
    Q_ASSERT(enumerator);
    // filter for mouse devices
    udev_enumerate_add_match_subsystem(enumerator, "input");
    udev_enumerate_add_match_property(enumerator, "ID_INPUT_MOUSE", "1");
    udev_enumerate_scan_devices(enumerator);
    udev_list_entry *device_entries =
        udev_enumerate_get_list_entry(enumerator);
    // iterate over all mouse devices
    udev_list_entry *current_entry = 0;
    udev_list_entry_foreach(current_entry, device_entries) {
        QUDevDevicePtr device = udev_device_new_from_syspath(
            d->context, udev_list_entry_get_name(current_entry));
        // a real mouse device
        if (device && isMouse(device)) {
            QByteArray serial = getSerial(device);
            // update the serial cache
            QByteArray sysPath = udev_device_get_syspath(device);
            d->serialCache.insert(serial, sysPath);
            kDebug() << "device with serial" << serial;
            devices << QString::fromAscii(serial);
        }
    }
    return devices;
}

QString MouseDevicesMonitor::productName(const QString &id) const {
    Q_D(const MouseDevicesMonitor);
    kDebug() << "querying product name for serial" << id;
    QByteArray serial = id.toAscii();
    // try to get product name from cache
    QByteArray name = d->productNameCache.value(serial);
    if (!name.isEmpty()) {
        kDebug() << "found product name" << name
                 << "for serial" << serial << "in cache";
        return QString::fromAscii(name);
    }
    // cache miss, query udev for the product name
    QUDevDevicePtr device = d->findDeviceBySerial(id.toAscii());
    if (device) {
        // the parent object is attached to the device object, and is not
        // referenced itself.  It is destroyed, when the device object is
        // destroyed, therefore no guarded pointer must be used to avoid
        // double frees.
        udev_device *parent = udev_device_get_parent(device);
        if (parent) {
            kDebug() << "found" << name << "for serial" << id;
            QByteArray name = udev_device_get_property_value(
                parent, "NAME");
            // cache the name
            d->productNameCache.insert(serial, name);
            return QString::fromAscii(name);
        }
    }
    return QString();
}

#include "moc_mousedevicesmonitor.cpp"
