/*
 * synaptiks -- a touchpad control tool
 *
 *
 * Copyright (C) 2010 Sebastian Wiesner <lunaryorn@googlemail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef QUDEVPOINTER_H
#define QUDEVPOINTER_H

extern "C" {
#include <libudev.h>
}

/**
 * @file   qudevpointer.h
 *
 * Provide a guarded pointer class for libudev types, and some convenient
 * typedefs for guarded pointers to common libudev types.
 */


/**
 * @brief A managed pointer a-top of libudev
 *
 * @param T the pointer type
 * @param Ref a function which increases the reference count on T*
 * @param Unref a function which decreases the reference count on T*
 */
template<class T, T* (*Ref)(T*), void (*Unref)(T*)> class QUDevPointer {
public:
    /**
     * @brief Default constructor
     */
    QUDevPointer(): m_ptr(0) {
    }

    /**
     * @brief Copy constructor.
     *
     * Copy internal pointer of @p other to this object and increase ref
     * count.
     */
    QUDevPointer(const QUDevPointer<T, Ref, Unref> &other) {
        this->m_ptr = other.m_ptr;
        Ref(this->m_ptr);
    }

    /**
     * @brief Construct a managed udev pointer from a raw pointer.
     *
     * The reference count of @p ptr is @em not increased.
     *
     * @param ptr the raw pointer
     */
    QUDevPointer(T *ptr) {
        this->m_ptr = ptr;
    }

    /**
     * @brief Destroy this pointer and decrease reference count on internal
     * pointer.
     */
    ~QUDevPointer() {
        Unref(this->m_ptr);
    }

    /**
     * @brief Assignment operator.
     *
     * Decrease reference count on the current internal pointer of this
     * object (if it has one), copy the internal pointer of @p other into
     * this object, and increase the reference count on the new internal
     * pointer.
     *
     * @param other managed pointer to assign to this pointer
     */
    QUDevPointer &operator=(const QUDevPointer<T, Ref, Unref> &other) {
        if (this->m_ptr)
            Unref(this->m_ptr);
        this->m_ptr = other->m_ptr;
        Ref(this->m_ptr);
        return *this;
    }

    /**
     * Decrease reference count on the current internal pointer of this
     * object (if it has one) and set internal pointer to @p ptr.
     *
     * @param ptr the raw pointer to assign to this pointer
     */
    QUDevPointer &operator=(T *ptr) {
        if (this->m_ptr)
            Unref(this->m_ptr);
        this->m_ptr = ptr;
        return *this;
    }

    /**
     * @brief Forcibly increase the reference count on the underlying
     * pointer.
     */
    void ref() {
        Ref(this->m_ptr);
    }

    /**
     * @brief Conversion operator to T*
     *
     * A managed udev pointer implicitly converts to T*.
     *
     * @return the underlying raw pointer
     */
    operator T*() const {
        return this->m_ptr;
    }

    /**
     * @brief Get the underyling raw pointer.
     *
     * @return the underlying raw pointer.
     */
    T *data() const {
        return this->m_ptr;
    }

    /**
     * @brief Implicit conversion to bool.
     *
     * @return @c true, if the internal pointer is not 0, @c false
     * otherwise.
     */
    operator bool() const {
        return this->m_ptr;
    }

    /**
     * @brief Test, if this pointer is 0.
     *
     * @return @c true, if the internal pointer is not 0, @c false
     *         otherwise.
     */
    bool isNull() const {
        return !this->m_ptr;
    }

private:
    /** internal pointer */
    T *m_ptr;
};

/** Managed pointer to @c udev_context* */
typedef QUDevPointer<struct udev, udev_ref, udev_unref> QUDevContextPtr;
/** Managed pointer to @c udev_enumerate* */
typedef QUDevPointer<struct udev_enumerate, udev_enumerate_ref,
                     udev_enumerate_unref> QUDevEnumeratePtr;
/** Managed pointer to @c udev_monitor* */
typedef QUDevPointer<struct udev_monitor, udev_monitor_ref,
                     udev_monitor_unref> QUDevMonitorPtr;
/** Managed pointer to @c udev_device* */
typedef QUDevPointer<struct udev_device, udev_device_ref,
                     udev_device_unref> QUDevDevicePtr;

#endif /* QUDEVPOINTER_H */
